#!/usr/bin/env python
# -*- coding: utf8 -*-

# Copyright 2022 Przemyslaw Bereski https://github.com/przemobe/

import ntw
import time
import struct
import binascii

# from datetime import date
# (date(2000, 1, 1) - date(1900, 1, 1)).days * 24*60*60
NTP_EPOCH_DELTA2000 = 3155673600
NTP_EPOCH_DELTA1970 = 2208988800
NTP_EPOCH_JST_P9H = 9 * 60 * 60

class SntpClient:
    '''Very simple SNTP client'''

    def __init__(self, ntw, client_port, server_addr, server_port=123):
        self.ntw = ntw
        self.client_port = client_port
        self.server_addr = bytes(server_addr)
        self.server_port = server_port
        # Define states: 0 - idle, 1 - connecting, 2 - await response, 3 - done
        self.state = 0
        self.rcnt = 0
        self.init_time = 0

        # Register callback for response
        self.ntw.registerUdp4Callback(self.client_port, self.proc_response)

    def loop(self):
        ctime = time.time()

        # State - idle
        if 0 == self.state:
            if not self.ntw.isIPv4Configured():
                return
            print('[SNTP] Connecting...')
            self.ntw.connectIp4(self.server_addr)
            self.init_time = ctime
            self.state = 1
            #self.ntw.rxARPPkt()
            # State - connecting
        elif 1 == self.state:
            if self.ntw.isConnectedIp4(self.server_addr):
                print('[SNTP] Connected')
                self.init_time = ctime
                self.state = 2
                if 0 > self.send_request():
                    self.state = 3
                    # Deregister callback for response
                    self.ntw.registerUdp4Callback(self.client_port, None)
            elif ctime - self.init_time > 3:
                # Retry
                self.state = 0
                self.rcnt += 1
                if self.rcnt >= 10:
                    print("Not Conected !!")
                    self.state = 3

        # State - await response
        elif 2 == self.state:
            if ctime - self.init_time > 3:
                # Retry
                self.state = 1

        # State - done
        else: # 3 == self.state
            pass

    def send_request(self):
        request = b'\x1b' + 47 * b'\0'
        print("SNTP Request=", binascii.b2a_hex(request,b' '))
        n = self.ntw.sendUdp4(self.server_addr, self.server_port, request, self.client_port)
        if 0 > n:
            print(f'[SNTP] Fail to send request: error={n}')
        return n

    def proc_response(self, pkt):
        if 2 != self.state:
            # Ignore unexpected pkt
            return

        self.state = 3
        t = struct.unpack('!12I', pkt.udp_data)[10]
        t -= NTP_EPOCH_DELTA1970
        t += NTP_EPOCH_JST_P9H
        print(f'[SNTP] Response received: time=' , time.strftime("%Y-%m-%d %H:%M:%S (JST)",time.gmtime(t)))

        # Deregister callback for response
        self.ntw.registerUdp4Callback(self.client_port, None)


if __name__ == '__main__':
    # Create network
    ntw = ntw.Ntw()

    # Set static IP address
    ntw.setIPv4([192,168,1,39], [255,255,255,0], [192,168,1,1])

    # Create SNTP client
    # NTP server address 'pool.ntp.org' is 162.159.200.1
    # Select unused local UDP port: 51000
    sntp_cli = SntpClient(ntw, 8123, [162,159,200,1])

    while True:
        ntw.rxAllPkt()
        sntp_cli.loop()
        if sntp_cli.state == 3:
            break

'''
$ python sntp.py 
..... INIT() exit .....
. HwRevID: 0x04
. Cntrl: ECON1 ECON2 ESTAT  EIR  EIE
.        0x04  0x80  0x01  0x00  0xD0
. MAC  : MACON1 MACON3 MACON4
.        0x0D   0xB3   0x40
. Rx   : ERXST  ERXND  ERXWRPT ERXRDPT ERXFCON EPKTCNT MAMXFL
.        0x0000 0x17FF 0x0000  0x17FF  0xA1    0x00    0x05EE
. Tx   : ETXST  ETXND  MACLCON1 MACLCON2 MAPHSUP
.        0x1800 0x1800 0x0F     0x37     0x10
..........
MAC ADDR: 3e:3f:3f:02:12:34
ENC28J60 revision ID: 0x04
[SNTP] Connecting...
ARP Request= b'ff ff ff ff ff ff 3e 3f 3f 02 12 34 08 06 00 01 08 00 06 04 00 01 3e 3f 3f 02 12 34 c0 a8 01 27 00 00 00 00 00 00 c0 a8 01 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00'
RECV= b'3e 3f 3f 02 12 34 XX xX XX X6 ed d5 08 06 00 01 08 00 06 04 00 02 XX xX XX X6 ed d5 c0 a8 01 01 3e 3f 3f 02 12 34 c0 a8 01 27 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00'
Rx ARP oper=2
ARP Register: 192.168.1.1 is at XX:XX:XX:X6:ED:D5
[SNTP] Connected
SNTP Request= b'1b 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00'
SEND NTP request: b'XX xX XX X6 ed d5 3e 3f 3f 02 12 34 08 00 45 00 00 4c 00 00 00 00 80 11 0e 31 c0 a8 01 27 a2 9f c8 01 1f bb 00 7b 00 38 97 d7 1b 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00'
RECV= b'3e 3f 3f 02 12 34 XX xX XX X6 ed d5 08 00 45 00 00 4c 7f bb 40 00 38 11 96 75 a2 9f c8 01 c0 a8 01 27 00 7b 1f bb 00 38 1f da 1c 03 00 e7 00 00 21 70 00 00 00 2f 0a 98 08 05 e8 9c 74 55 70 d5 43 24 00 00 00 00 00 00 00 00 e8 9c 74 a3 02 c4 c5 97 e8 9c 74 a3 02 c9 ab 45'
Rx my IP proto=17
[SNTP] Response received: time= 2023-09-01 23:33:07 (JST)
'''
